/*
 * filter.c: URL filter daemon
 *
 * Copyright (C) 2021-2024 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Jomily Joseph <jomily.joseph@iopsys.eu>
 * Author: Rahul Thakur <rahul.thakur@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <arpa/inet.h>
#include <linux/netfilter.h>
#include <syslog.h>
#include <libnetfilter_queue/libnetfilter_queue.h>
#include <libnetfilter_queue/linux_nfnetlink_queue.h>
#include <libubox/uloop.h>
#include <pthread.h>

#include "filter.h"
#include "filter_utils.h"
#include "ubus.h"
#include "utils.h"
#include "dns_cache.h"

struct nfq_handle *nfq_h;
struct nfq_q_handle *queue_h;
struct uloop_fd uloop_fd;

// Checks if the packet needs to be accepted or not based on mac address
// and url.
// @param payload packet payload
// @return PKT_ACCEPT or PKT_DROP
static int handle_packet_filtering(struct nfq_data *payload)
{
	struct nfqnl_msg_packet_hw *hw_hdr;
	unsigned char *data = NULL;
	char payload_pkt_arr[INET6_ADDRSTRLEN] = {0};
	char *payload_pkt = payload_pkt_arr;
	uint8_t proto;
	size_t data_len;
	int ret = 0;
	bool is_https_packet = false;
	bool is_ipv6 = false;

	// Get the source mac address.
	hw_hdr = nfq_get_packet_hw(payload);
	if (!hw_hdr) {
		syslog(LOG_ERR, "Error in fetching mac");
		return PKT_ACCEPT;
	}

	char mac_addr[MAC_LEN];

	memset(mac_addr, '\0', MAC_LEN);
	snprintf(mac_addr, sizeof(mac_addr), "%02x:%02x:%02x:%02x:%02x:%02x",
			hw_hdr->hw_addr[0], hw_hdr->hw_addr[1], hw_hdr->hw_addr[2],
			hw_hdr->hw_addr[3], hw_hdr->hw_addr[4], hw_hdr->hw_addr[5]);

	if (mac_addr[0] == '\0') {
		syslog(LOG_ERR, "MAC address empty");
		return PKT_ACCEPT;
	}

	ret = nfq_get_payload(payload, &data);
	if (ret <= 0) {
		syslog(LOG_ERR, "%s:%u-->No payload found", __func__, __LINE__);
		return PKT_ACCEPT;
	}

	data_len = (size_t)ret;

	// Get the payload offset.
	int payload_offset;

	payload_offset = get_ip_packet_payload(data, &proto, data_len, &is_ipv6);
	if (payload_offset <= 0) {
		// Just accept the packet and return since the packet either does not match
		// urlfilter criteria or is a dns response.
		return PKT_ACCEPT;
	}

	ret = get_payload_pkt(proto, data, payload_offset,
			&payload_pkt, &is_https_packet, is_ipv6);

	if (ret != URL_SUCCESS)
		return ret;

	size_t len = strlen(payload_pkt);

	if (len == 0) {
		syslog(LOG_ERR, "payload is empty");
		return PKT_ACCEPT;
	}

	// Check if the URL has been blacklisted in any profile.
	ret = profile_based_url_filtering(payload_pkt, mac_addr, sizeof(mac_addr), is_https_packet, is_ipv6);

	return ret;
}


// Callback function for handling packets.
// @param queue_h queue 0 handle
// @param nfmsg nfq message received
// @param nfa message data
// @param data store data
// @return 0 on success and non zero otherwise
static int handle_packets(struct nfq_q_handle *q_h, struct nfgenmsg *nfmsg,
		struct nfq_data *nfa, void *data)
{
	struct nfqnl_msg_packet_hdr *pkt_hdr;
	int pkt_decision;
	uint32_t pkt_id = 0;

	// Fetch the packet header.
	pkt_hdr = nfq_get_msg_packet_hdr(nfa);
	if (pkt_hdr)
		pkt_id = ntohl(pkt_hdr->packet_id);

	// Check if we should block this packet on the basis of URL and source mac address.
	pkt_decision = handle_packet_filtering(nfa);
	if (pkt_decision == PKT_ACCEPT)
		return nfq_set_verdict(q_h, pkt_id, NF_ACCEPT, 0, NULL);
	else
		return nfq_set_verdict(q_h, pkt_id, NF_DROP, 0, NULL);
}

// Function to unbind and bind to nf_queue handler.
// @return nothing
static void update_nf_queue_handler(void)
{
	int nfq_bind_v4 = 1, nfq_bind_v6 = 1;

	// Unbinding existing nf_queue handler.
	if (nfq_unbind_pf(nfq_h, AF_INET) < 0) {
		syslog(LOG_ERR, "Error during nfq_unbind_pf() for IPv4");
		nfq_bind_v4 = 0;
	}

	if (nfq_unbind_pf(nfq_h, AF_INET6) < 0) {
		syslog(LOG_ERR, "Error during nfq_unbind_pf() for IPv6");
		nfq_bind_v6 = 0;
	}

	if (!(nfq_bind_v4 || nfq_bind_v6)) {
		syslog(LOG_ERR, "Error during nfq_unbind_pf()");
		exit(1);
	}

	// Binding nfnetlink_queue as nf_queue handler.
	nfq_bind_v4 = nfq_bind_v6 = 1;
	if (nfq_bind_pf(nfq_h, AF_INET < 0)) {
		syslog(LOG_ERR, "Error during nfq_bind_pf() for IPv4");
		nfq_bind_v4 = 0;
	}

	if (nfq_bind_pf(nfq_h, AF_INET6) < 0) {
		syslog(LOG_ERR, "Error during nfq_bind_pf() for IPv6");
		nfq_bind_v6 = 0;
	}

	if (!(nfq_bind_v4 || nfq_bind_v6)) {
		syslog(LOG_ERR, "Error during nfq_bind_pf()");
		exit(1);
	}
}

// Open a netlink connection and returns file descriptor
// @return fd of the netlink socket
static int open_netlink_connection(void)
{
	// Opening netlink socket.
	nfq_h = nfq_open();
	if (!nfq_h) {
		syslog(LOG_ERR, "Error during nfq_open()");
		exit(1);
	}

	update_nf_queue_handler();

	// Binding this socket to queue.
	queue_h = nfq_create_queue(nfq_h,  0, &handle_packets, NULL);
	if (!queue_h) {
		syslog(LOG_ERR, "Error during nfq_create_queue()");
		exit(1);
	}

	// Setting copy_packet mode.
	if (nfq_set_mode(queue_h, NFQNL_COPY_PACKET, 0xffff) < 0) {
		syslog(LOG_ERR, "nfq_set_mode error");
		exit(1);
	}

	return nfq_fd(nfq_h);
}

// This function prints the usage of the daemon
static void usage(char *prog)
{
	fprintf(stderr, "Usage: %s [options]\n", prog);
	fprintf(stderr, "\n");
	fprintf(stderr, "options:\n");
	fprintf(stderr, "    -s <socket path>   ubus socket\n");
	fprintf(stderr, "    -h <help>       To print this help menu\n");
	fprintf(stderr, "\n");
}

// This function reads the packets received on the fd
// @param fd is the pointer to the file descriptor to read on
static void url_filter_recv(struct uloop_fd *fd, unsigned int events)
{
	char buf[BUFSIZE];

	(void)events;

	memset(buf, 0, sizeof(buf));
	int rv = (int)recv(fd->fd, buf, sizeof(buf), 0);
	if (rv >= 0) {
		nfq_handle_packet(nfq_h, buf, rv);
	}
}

int main(int argc, char **argv)
{
	const char *socket = NULL;
	int ch;

	while ((ch = getopt(argc, argv, "hs:")) != -1) {
		switch (ch) {
		case 's':
			socket = optarg;
			break;
		case 'h':
			usage(argv[0]);
			exit(0);
		default:
			break;
				}
	}

	openlog("urlfilter", LOG_CONS | LOG_PID | LOG_NDELAY, LOG_LOCAL1);

	if (urlfilter_ubus_init(socket) < 0) {
		fprintf(stderr, "Failed to connect to ubus\n");
		goto exit;
	}

	// Call function to update the global structures.
	if (read_config()) {
		syslog(LOG_ERR, "failed to read config");
		goto exit;
	}

	// Create a thread which would send out dns query
	// for all the blacklisted urls to create a dns cache.
	pthread_t ptid;

	if (pthread_create(&ptid, NULL, &send_dns_queries, NULL) != 0) {
		syslog(LOG_ERR, "pthread create failed");
		goto exit;
	}

	uloop_fd.fd = open_netlink_connection();
	uloop_fd.cb = url_filter_recv;
	if (uloop_fd_add(&uloop_fd, ULOOP_READ)) {
		syslog(LOG_ERR, "url fitler: uloop_fd_add(%d) failed", uloop_fd.fd);
		goto destroy;
	}

	// Main loop of urlfilter
	uloop_run();

destroy:
	// Destroying the queue handle and closing nfq handle.
	nfq_destroy_queue(queue_h);
	nfq_close(nfq_h);

exit:
	free_all();  // Free all allocated resources
	urlfilter_ubus_done();
	uloop_done();

	return 0;
}
