/*
 * filter_utils.c: URL filter utils
 *
 * Copyright (C) 2024 iopsys Software Solutions AB. All rights reserved.
 *
 * Author: Rahul Thakur <rahul.thakur@iopsys.eu>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * version 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <arpa/inet.h>
#include <netinet/ip.h>
#include <netinet/ip6.h>
#include <netinet/tcp.h>
#include <netinet/udp.h>
#include <stdbool.h>
#include <syslog.h>
#include <ctype.h>
#include <regex.h>

#include "dns.h"
#include "config.h"
#include "utils.h"
#include "filter_utils.h"

static size_t ipv6_header_get_L4_header_offset(const struct ip6_hdr *ip6h, uint8_t *nexthdr_p)
{
	unsigned int ext_head_count = 8;
	const struct ip6_ext *ip_ext_p = (const struct ip6_ext *)(ip6h + 1);
	unsigned int payload_offset = sizeof(struct ip6_hdr);

	*nexthdr_p = ip6h->ip6_nxt;
	do {
		if ((*nexthdr_p == IPPROTO_TCP) || (*nexthdr_p == IPPROTO_UDP))
			break;

		payload_offset += (ip_ext_p->ip6e_len + 1U) << 3U;
		*nexthdr_p = ip_ext_p->ip6e_nxt;
		ip_ext_p = (const struct ip6_ext *)((const uint8_t *)ip6h + payload_offset);
		ext_head_count--; /* at most 8 extension headers */
	} while (ext_head_count);

	return payload_offset;
}

// Check if the packet is https or not. If yes then,
// returns the destination IP address.
// @param data payload data
// @return On success destination ip address else NULL
static char *check_https_packet(unsigned char *data, char **payload_pkt, bool is_ipv6)
{
	struct sockaddr_in dest;
	const struct iphdr *iph;
	const struct tcphdr *tcp_header;
	uint8_t proto;
	const struct ip6_hdr *ip6h;
	struct sockaddr_in6 dest6;

	if (!is_ipv6) {
		iph = (const struct iphdr *)data;
		tcp_header = (const struct tcphdr *)(data + (iph->ihl<<2));

		memset(&dest, 0, sizeof(dest));
		dest.sin_addr.s_addr = iph->daddr;

		// Check if destination port is 443.
		if (ntohs(tcp_header->dest) == HTTPS_PORT)
			return inet_ntoa(dest.sin_addr);
	} else {
		ip6h = (const struct ip6_hdr *)data;
		tcp_header = (const struct tcphdr *)(data + ipv6_header_get_L4_header_offset((struct ip6_hdr *)data, &proto));

		memset(&dest6, 0, sizeof(dest6));
		memcpy(&dest6.sin6_addr, &ip6h->ip6_dst, sizeof(struct in6_addr));

		if (ntohs(tcp_header->dest) == HTTPS_PORT)
			return (char *)inet_ntop(AF_INET6, (const struct sockaddr_in6 *)&dest6.sin6_addr, *payload_pkt, INET6_ADDRSTRLEN);
	}

	return NULL;
}

static int handle_udp_packet(const unsigned char *data_p, size_t offset, size_t len)
{
	const struct udphdr *udp_header = (const struct udphdr *)(data_p + offset);

	if (((ntohs(udp_header->dest)) != DNS_PORT) && ((ntohs(udp_header->source)) != DNS_PORT)) {
		syslog(LOG_ERR, "Not a DNS packet");
		return -1;
	}

	if (ntohs(udp_header->dest) == DNS_PORT) {
		return 0;
	}

	if (ntohs(udp_header->source) == DNS_PORT) {
		unsigned char dns_payload[len + 1];

		memset(dns_payload, '\0', sizeof(dns_payload));
		memcpy(dns_payload, data_p + offset + sizeof(struct udphdr), len - (offset + sizeof(struct udphdr)) + 1);
		if (dns_payload[0] == '\0') {
			syslog(LOG_ERR, "%s:%u--> payload is null", __func__, __LINE__);
			return -1;
		}
		parse_dns_response(dns_payload);
		return 0;
	}

	return (int)(offset + sizeof(struct udphdr) + sizeof(struct dnshdr));
}

static int handle_tcp_packet(const unsigned char *data_p, size_t offset, size_t len)
{
	const struct tcphdr *tcp_header = (const struct tcphdr *)(data_p + offset);
	size_t tcp_header_length = tcp_header->doff * 4;

	if (ntohs(tcp_header->source) == DNS_PORT) {
		unsigned char dns_payload[len + 1];

		memset(dns_payload, '\0', sizeof(dns_payload));
		if (len <= (offset + tcp_header_length)) {
			return -1;
		}

		memcpy(dns_payload, data_p + offset + tcp_header_length, len - (offset + tcp_header_length) + 1);
		if (dns_payload[0] == '\0') {
			syslog(LOG_ERR, "payload is null");
			return -1;
		}

		parse_dns_response(dns_payload);
		return 0;
	} else if (ntohs(tcp_header->dest) == DNS_PORT) {
		return 0;
	}

	return (int)(offset + tcp_header_length);
}

static int process_ipv4_packet(const unsigned char *data_p, uint8_t *proto, size_t len, bool *is_ipv6) {
	const struct iphdr *iph = (const struct iphdr *)data_p;
	size_t ip_header_length = iph->ihl * 4;
	*proto = iph->protocol;
	*is_ipv6 = false;

	if (*proto == IPPROTO_UDP) {
		return handle_udp_packet(data_p, ip_header_length, len);
	} else if (*proto == IPPROTO_TCP) {
		return handle_tcp_packet(data_p, ip_header_length, len);
	} else {
		syslog(LOG_ERR, "Unknown protocol");
		return -1;
	}
}

static int process_ipv6_packet(const unsigned char *data_p, uint8_t *proto, size_t len, bool *is_ipv6)
{
	const struct ip6_hdr *ip6h = (const struct ip6_hdr *)data_p;

	if ((ip6h->ip6_vfc >> 4) != 6) {
		return -1;
	}

	*is_ipv6 = true;

	size_t payload_offset = ipv6_header_get_L4_header_offset(ip6h, proto);
	if (*proto == IPPROTO_UDP) {
		return handle_udp_packet(data_p, payload_offset, len);
	} else if (*proto == IPPROTO_TCP) {
		return handle_tcp_packet(data_p, payload_offset, len);
	} else {
		syslog(LOG_ERR, "Unknown protocol");
		return -1;
	}
}

int get_ip_packet_payload(unsigned char *data_p, uint8_t *proto, size_t len, bool *is_ipv6)
{
	const struct iphdr *iph = (const struct iphdr *)data_p;
	bool ipv4 = (iph->version == 4);

	if (ipv4) {
		return process_ipv4_packet(data_p, proto, len, is_ipv6);
	} else {
		return process_ipv6_packet(data_p, proto, len, is_ipv6);
	}
}

// Function fetches the packet payload ie URL for http packet,
// destination IP address for https packet and domain name for
// dns packet
// @param proto protocol type
// @param data payload
// @param payload_offset data offset
// @param payload_pkt variable to store payload
// @param is_https_packet variable to store if its a https packet
// @return 1 or PKT_ACCEPT
int get_payload_pkt(int proto, unsigned char *data, int payload_offset,
		char **payload_pkt, bool *is_https_packet, bool is_ipv6)
{
	if (proto == IPPROTO_TCP) {
		// If the packet is a dns reponse then accept the packet.
		if (payload_offset == 0)
			return PKT_ACCEPT;
		// Check if its https packet or not, if yes then update the
		// payload_pkt with destination ip address.
		*payload_pkt = check_https_packet(data, payload_pkt, is_ipv6);

		if (*payload_pkt != NULL && *payload_pkt[0] != '\0') {
			*is_https_packet = true;
			return URL_SUCCESS;
		}

		// Since its not an https packet, check if it is an http packet or not
		*payload_pkt = (char *)(data + payload_offset);
		if (strstr(*payload_pkt, "GET ") == NULL  &&
				strstr(*payload_pkt, "POST ") == NULL &&
				strstr(*payload_pkt, "HEAD ") == NULL) {
			syslog(LOG_INFO, "No http/https packet");
			return PKT_ACCEPT;
		}
	} else if (proto == IPPROTO_UDP) {
		// If the packet is a dns reponse then accept the packet.
		if (payload_offset == 0)
			return PKT_ACCEPT;

		// Get the qname(hostname) from the packet.
		*payload_pkt = get_domain_name((char *)(data + payload_offset));
		if (*payload_pkt == NULL) {
			syslog(LOG_INFO, "Payload for dns packet is NULL");
			return PKT_ACCEPT;
		}
	}

	return URL_SUCCESS;
}

static int match_url_list(struct list_head *url_list, char *payload_pkt)
{
	struct url_list *url;
	regex_t regex;
	int reti;
	char url_regex[1024] = {0};

	list_for_each_entry(url, url_list, list) {
		// Check if url->url is a substring of payload_pkt
		if (strstr(payload_pkt, url->url) != NULL) {
			url->match_cnt++;
			return PKT_DROP;
		}

		// Convert wildcard to regex and compile the regular expression
		wildcard_to_regex(url->url, url_regex);
		reti = regcomp(&regex, url_regex, REG_ICASE | REG_EXTENDED);
		if (reti) {
			syslog(LOG_ERR, "Could not compile regex: %s", url_regex);
			continue;
		}

		// Execute the regular expression to check the payload against the URL list
		reti = regexec(&regex, payload_pkt, 0, NULL, 0);
		regfree(&regex);

		if (!reti) {
			url->match_cnt++;
			return PKT_DROP;
		}
	}

	return CHECK_CONTINUE;
}

// Function checks if the URL requested is in blacklist or whitelist for a
// particular profile.
// @param payload_pkt packet payload
// @param match_profile profile store which has the requested mac address.
// @return PKT_ACCEPT or PKT DROP on success else -1.
static int match_url_from_filter(char *payload_pkt, struct urlfilter *filter)
{
	int result;

	if (!filter) {
		syslog(LOG_INFO, "filter or bundle is null.");
		return CHECK_CONTINUE;
	}

	// If access_policy is true, return PKT_ACCEPT right away
	if (filter->access_policy) {
		return PKT_ACCEPT;
	}

	// Check URLs in filter_text list
	result = match_url_list(&filter->filter_text, payload_pkt);
	if (result != CHECK_CONTINUE) {
		return result;
	}

	// Check URLs from the urlbundle
	if (filter->bundle) {
		result = match_url_list(&filter->bundle->custom_url, payload_pkt);
		if (result != CHECK_CONTINUE) {
			return result;
		}

		// Check blacklist URLs (download_url) from the urlbundle
		result = match_url_list(&filter->bundle->download_url, payload_pkt);
		if (result != CHECK_CONTINUE) {
			return result;
		}
	}

	return CHECK_CONTINUE;
}


// Function checks if the current day and time is within
// the given constraint or not.
// @param day_index current day index.
// @param prev_day_index previous day index
// @param time_data structure storing data.
// @return 0 on success else -1.
static int check_if_time_within_range(struct schedule *schedule, unsigned int day_index,
					unsigned int prev_day_index, time_info time_data)
{
	int j = 0;
	bool is_day_present = false;
	struct schedule *sched;

	// Ensure filter and schedule are valid
	if (!schedule || !schedule->enable) {
		syslog(LOG_INFO, "schedule is null or not enabled.");
		return CHECK_CONTINUE;
	}

	sched = schedule;

	// Check if the day is specified in the configuration.
	// If no day is specified (i.e., no active days), the filter is applied every day.
	for (j = 0; j < MAX_DAYS; j++) {
		if (sched->day[j] == FILTER_ACTIVE) {
			is_day_present = true;
			break;
		}
	}

	// If no specific day is active, apply the filter for the specified time and duration.
	if (!is_day_present) {
		if ((time_data.current_time >= time_data.start_time) &&
				(time_data.current_time <= time_data.end_time))
			return 0;

		// If the filter spans into the next day, check the next day's end time.
		if (time_data.is_next_day) {
			if (time_data.current_time <= time_data.next_day_end_time)
				return 0;
		}

		return CHECK_CONTINUE;
	}

	// Now check if the filter is active for the current day.
	if (sched->day[day_index] == FILTER_ACTIVE) {
		// Check if the current time is within the time range for the current day.
		if ((time_data.current_time >= time_data.start_time) &&
				(time_data.current_time <= time_data.end_time))
			return 0;

		// If the filter spans into the next day and was active for the previous day, check that time range.
		if (time_data.is_next_day && sched->day[prev_day_index] == FILTER_ACTIVE) {
			if (time_data.current_time <= time_data.next_day_end_time)
				return 0;
		}
	} else {
		// If the filter is not active for the current day, check if it was active for the previous day
		// and spans into the next day.
		if (time_data.is_next_day && sched->day[prev_day_index] == FILTER_ACTIVE) {
			if (time_data.current_time <= time_data.next_day_end_time)
				return 0;
		}
	}

	return CHECK_CONTINUE;
}

// Function filters the packet based on time slots.
// @return 0 on success else -1.

static int time_based_url_filtering(struct schedule *schedule)
{
	time_info time_data;
	unsigned int day_index = 0, prev_day_index = 0,
		     duration = 0, hh = 0, ss = 0, mm = 0;

	if (!schedule || !schedule->enable) {
		syslog(LOG_INFO, "schedule is null or not enabled.");
		return 0;  // Return 0 to indicate no filtering is applied
	}

	memset(&time_data, 0, sizeof(time_info));

	// Fetch the current day and time.
	time_data.current_time = get_current_day_and_time(&day_index);

	// Get previous day index
	prev_day_index = day_index ? day_index - 1 : MAX_DAYS;

	// Check if duration is greater than 24 hrs (86400 seconds); if so, set duration to 24 hrs.
	duration = schedule->duration;
	if (duration >= MAX_SEC)
		duration = MAX_SEC;

	// Convert start_time into seconds and add duration to get the end time.
	// If the end time exceeds 86400 seconds, set is_next_day flag and calculate the
	// next day's end time.
	sscanf(schedule->start_time, "%u:%u:%u", &hh, &mm, &ss);
	time_data.start_time = ((hh * 60) + mm) * 60 + ss;
	time_data.end_time = time_data.start_time + duration;

	if (time_data.end_time > MAX_SEC) {
		time_data.is_next_day = 1;
		time_data.next_day_end_time = time_data.end_time - MAX_SEC;
		time_data.end_time = MAX_SEC;
	}

	// Check if the current time falls within the schedule range.
	return check_if_time_within_range(schedule, day_index, prev_day_index, time_data);
}

// Function checks if the requested URL is present in the profile.
// If yes, then either accepts it or drops it as per the configuration.
// @param payload_pkt packet data
// @param match_profile profile which matches with the source mac address
// @param is_https_packet https packet or not
// @return on success PKT_ACCEPT or PKT_DROP else -1.
static int mac_based_url_filtering(char *payload_pkt, struct urlfilter *filter,
						bool is_https_packet, bool is_ipv6)
{
	int ret;
	char *url_pkt = NULL;

	// If the packet is https then find the URL corresponding to
	// the IP address present in the payload_pkt from the dns cache.
	if (is_https_packet == true) {
		url_pkt = match_ip_address_from_dns_cache(payload_pkt, is_ipv6);
		if (url_pkt == NULL)
			return PKT_ACCEPT;
		ret = match_url_from_filter(url_pkt, filter);
	} else {
		ret = match_url_from_filter(payload_pkt, filter);
		if (ret == CHECK_CONTINUE) {
			// Check if the url is present in cname list of dns cache.
			// If yes, then find the url name corresponding to the cname and check
			// if that is present in the matched profile. If yes, then drop the packet.
			char *cname = match_cname_from_dns_cache(payload_pkt);
			if (cname != NULL)
				ret = match_url_from_filter(cname, filter);
		}
	}

	return ret;
}

// Function finds the profile associated with the MAC address of the
// packet received and maps the URL with the requested URL.
// @param mac_addr mac address to be searched in the profile
// @param len length of mac_addr
// @return on success PKT_ACCEPT or PKT_DROP else -1.
int profile_based_url_filtering(char *payload_pkt, char *mac_addr,
					size_t len, bool is_https_packet, bool is_ipv6)
{
	int ret;
	struct profile *p;
	struct urlfilter *filter;
	struct host *h;
	bool host_found = false;

	// Iterate over the global profile list
	list_for_each_entry(p, &profiles, list) {
		// Iterate over the hosts (MAC addresses) associated with this profile
		list_for_each_entry(h, &p->hosts, list) {
			if (strncasecmp(mac_addr, h->host_identifier, len) == 0) {
				host_found = true;
				break;
			}
		}

		if (host_found) {
			// Iterate over the filters associated with the profile
			list_for_each_entry(filter, &p->filters, list) {
				// Check if the filter is enabled
				if (!filter->enable) {
					continue;
				}

				// Check if the filter has an active schedule first
				struct schedule *schedule = NULL;
				bool schedule_active = false;

				// no schedule means filter is active at all times
				if (list_empty(&filter->schedules)) {
					// The list is empty, no elements
					schedule_active = true;
				} else {
					// The list contains one or more elements
					list_for_each_entry(schedule, &filter->schedules, list) {
						// Perform time-based filtering first
						if (!time_based_url_filtering(schedule)) {
							schedule_active = true;
							break;
						}
					}
				}

				// If no schedule is active, skip URL filtering
				if (!schedule_active) {
					continue;
				}

				// Perform URL filtering if the schedule is active
				ret = mac_based_url_filtering(payload_pkt, filter, is_https_packet, is_ipv6);

				// If URL filtering results in a drop, return the result
				if (ret == PKT_DROP) {
					return ret;
				}

				// else continue on to the next filter
			}
		}
	}

	return PKT_ACCEPT;
}
